r"""
Conduit Lengths Models
----------------------

This model contains a selection of functions for Calculating conduit lengths
in the network assuming pores and throats have different shapes.

"""
from ._funcs import *
