r"""
Throat Capillary Shape Factor Models
------------------------------------

This model contains a selection of functions for calculating capillary shape
factor for throats.

"""

from ._funcs import *
