r"""
Collection of functions for manipulating network topology
=========================================================

This module contains a selection of functions that deal specifically with
network topology.

"""

from ._topotools import *
from ._perctools import *
from ._graphtools import *
